import {S3Client, GetObjectCommand} from "@aws-sdk/client-s3";
import {Upload} from '@aws-sdk/lib-storage';
import Sharp from 'sharp';
// import Stream from 'stream';

const PathPattern = /^([^\/]+)\/((?:[^\/]+\/)*)([^\/]+)$/;

// parameters
const {BUCKET, URL} = process.env;
const WHITELIST = process.env.WHITELIST
    ? Object.freeze(process.env.WHITELIST.split(' '))
    : null;

const s3Client = new S3Client();

export const handler = async (event) => {
    const path = event.queryStringParameters?.path;
    if (!path) {
        return errorResponse('Parameter "?path=" is empty or missing');
    }

    const parts = PathPattern.exec(path.split('?')[0]);
    if (!parts || !parts[1] || !parts[3]) {
        return errorResponse('Parameter "?path=" should look like: 150x150/path/image.jpg');
    }
    
    const resizeOption = parts[1];  // e.g. "150x150_max"
    const dir = parts[2] || '';
    const sizeAndAction = resizeOption.split('_');
    const filename = parts[3];

    const sizes = sizeAndAction[0].split("x");
    const action = sizeAndAction.length > 1 ? sizeAndAction[1] : null;

    // Whitelist validation.
    if (WHITELIST && !WHITELIST.includes(resizeOption)) {
        return errorResponse(`WHITELIST is set but does not contain the size parameter "${resizeOption}"`, 403);
    }

    // Action validation.
    if (action && action !== 'max' && action !== 'min') {
        return errorResponse(
            `Unknown func parameter "${action}"\n` +
            'For query ".../150x150_func", "_func" must be either empty, "_min" or "_max"'
        );
    }

    try {
        const originImage = await s3Client.send(new GetObjectCommand({
            Bucket: BUCKET,
            Key: dir + filename,
        }));

        const width = sizes[0] === 'AUTO' ? null : parseInt(sizes[0]);
        const height = sizes[1] === 'AUTO' ? null : parseInt(sizes[1]);
        let fit;
        switch (action) {
            case 'max':
                fit = 'inside';
                break;
            case 'min':
                fit = 'outside';
                break;
            default:
                fit = 'cover';
                break;
        }

        // 将 S3 获取的流缓冲为 buffer
        const imageBuffer = await streamToBuffer(originImage.Body);
        // 获取图像元数据
        const metadata = await Sharp(imageBuffer).metadata();

        // 水印图片尺寸是200x56, 为了确保加水印成功，设置最小宽度250，高度100
        if (Math.min(...[metadata.width, width].filter(x => x !== null)) >= 250 && Math.min(...[metadata.height, height].filter(x => x !== null)) >= 100) {
            var sharpBuffer = await Sharp(imageBuffer).composite([{
            input: './ebest_watermark.png',
            gravity: 'center',
            blend: 'over'
            }]).resize(width, height, { withoutEnlargement: true, fit }).rotate().toBuffer();
        } else {
            var sharpBuffer = await Sharp(imageBuffer).resize(width, height, { withoutEnlargement: true, fit }).rotate().toBuffer();
        }

        // This does not work: await s3Client.send(new PutObjectCommand({params})
        // Solution: https://github.com/aws/aws-sdk-js-v3/issues/1920#issuecomment-761298908
        const upload = new Upload({
            client: s3Client,
            params: {
                Bucket: BUCKET,
                Key: path,
                Body: sharpBuffer,
                ContentType: originImage.ContentType,
                CacheControl: 'public, max-age=86400'
            },
        });
        await upload.done();

        return {
            statusCode: 301,
            headers: {"Location" : `${URL}/${path}`}
        };
    } catch (e) {
        return errorResponse('Exception: ' + e.message, e.statusCode || 400);
    }
}

function errorResponse(body, statusCode = 400) {
    return {
        statusCode,
        body,
        headers: {"Content-Type": "text/plain"}
    }
}

function streamToBuffer(stream) {
    return new Promise((resolve, reject) => {
        const chunks = [];
        stream.on('data', chunk => chunks.push(chunk));
        stream.on('end', () => resolve(Buffer.concat(chunks)));
        stream.on('error', reject);
    });
}
